#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>


#define US2CLOCK(us) (((uint32_t)(us) * 60398U) >> 16)

/*
 *       
 *    . ,    
 *    ,   1   
 *  (  )
 */

static uint8_t PROGMEM pin_map[32] = {
	7, 3, 2, 6, 5, 1, 0, 4,
	7 + 8, 3 + 8, 2 + 8, 6 + 8, 5 + 8, 1 + 8, 0 + 8, 4 + 8,
	4 + 16, 0 + 16, 1 + 16, 5 + 16, 6 + 16, 2 + 16, 3 + 16, 7 + 16,
	4 + 24, 0 + 24, 1 + 24, 5 + 24, 6 + 24, 2 + 24, 3 + 24, 7 + 24,
};


static uint8_t current_servo[4] = {0,0,0,0};

static uint16_t pause[4] = {
	US2CLOCK(20500), US2CLOCK(20500), US2CLOCK(20500), US2CLOCK(20500)
};

static uint16_t servo_position[32] = {
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),	US2CLOCK(1500),
};

void set_position(uint8_t n, uint16_t pos)
{
	if (n > 31) return;
	if (pos < 500) pos = 500;
	else if (pos > 2500) pos = 2500;
	servo_position[pgm_read_byte(pin_map+n)] = US2CLOCK(pos);
}

static void inline process_timer(volatile uint16_t* OCRX, volatile uint8_t* TCCRX, uint8_t block, uint16_t FOC_MASK) __attribute__((always_inline));

static void inline process_timer(volatile uint16_t* OCRX, volatile uint8_t* TCCRX, uint8_t block, uint16_t FOC_MASK)
{
	if (current_servo[block] == 8) {
		current_servo[block] = 0;
		*OCRX += pause[block];
		pause[block] = US2CLOCK(20500) ;
		*TCCRX |= FOC_MASK;
		*TCCRX |= FOC_MASK;
	} else {
		*OCRX += servo_position[block * 8 + current_servo[block]];
		pause[block] -= servo_position[block * 8 + current_servo[block]];
		current_servo[block]++;
	}
	*TCCRX |= FOC_MASK;
}

SIGNAL(SIG_OUTPUT_COMPARE3A)
{
	process_timer(
			(uint16_t*)_SFR_ADDR(OCR3A),
			(uint8_t*)_SFR_ADDR(TCCR3C),
		   	2,
			_BV(FOC3A)
	);
}

SIGNAL(SIG_OUTPUT_COMPARE3B)
{
	process_timer(
			(uint16_t*)_SFR_ADDR(OCR3B),
			(uint8_t*)_SFR_ADDR(TCCR3C),
		   	0,
			_BV(FOC3B)
	);
}

SIGNAL(SIG_OUTPUT_COMPARE3C)
{
	process_timer(
			(uint16_t*)_SFR_ADDR(OCR3C),
			(uint8_t*)_SFR_ADDR(TCCR3C),
		   	3,
			_BV(FOC3C)
	);
}

SIGNAL(SIG_OUTPUT_COMPARE1C)
{
	process_timer(
			(uint16_t*)_SFR_ADDR(OCR1C),
			(uint8_t*)_SFR_ADDR(TCCR1C),
		   	1,
			_BV(FOC1C)
	);
}

int main(void)
{
	/*        */
	DDRE = _BV(2) | _BV(3) | _BV(4) | _BV(5);
	DDRB = _BV(7);

	/*      */
	PORTE |= _BV(2);
	/*       */
	OCR3A = 2000;
	OCR3B = 2000;
	OCR3C = 2000;
	OCR1C = 2000;

	/*       0  0xFFFF.   
	 *   OCR3X   OC3X   . */
	TCCR3A = _BV(COM3A0) | _BV(COM3B0) | _BV(COM3C0);
	/*    = 8 */
	TCCR3B = _BV(CS31);
	/*       OCRXX */
	ETIMSK = _BV(OCIE3A);
	ETIMSK |= _BV(OCIE3B);
	ETIMSK |= _BV(OCIE3C);

	TCCR1A = _BV(COM1C0);
	/*    = 8 */
	TCCR1B = _BV(CS11);
	/*       OCR1C */
	ETIMSK |= _BV(OCIE1C);

	/*      */
	PORTE &= ~_BV(2);
	sei();					/*   */

	for(;;)  {
		
	}
}

